﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.DesktopEdition;
using ThinkGeo.MapSuite.Routing;
using System.Globalization;
using System.Collections.ObjectModel;
using System.Diagnostics;

namespace HelloRoutingWorld
{
    public partial class Form1 : Form
    {
        private RoutingEngine routingEngine;
        private RoutingSource routingSource;
        private RoutingLayer routingLayer;

        private InMemoryFeatureLayer pointLayer;

        
        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            winformsMap1.MapUnit = GeographyUnit.DecimalDegree;
            winformsMap1.CurrentExtent = new RectangleShape(12, 52, 19, 48);
            ShapeFileFeatureSource.BuildIndexFile(@"C:\CR\railways.shp", BuildIndexMode.DoNotRebuild);

            ShapeFileFeatureSource railsSource = new ShapeFileFeatureSource(@"C:\CR\railways.shp");
            ShapeFileFeatureSource roadsSource = new ShapeFileFeatureSource(@"C:\CR\roads.shp");
            ShapeFileFeatureLayer railsLayer = new ShapeFileFeatureLayer(@"C:\CR\railways.shp");

            railsLayer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = LineStyles.Railway3;
            railsLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            LayerOverlay networksOverlay = new LayerOverlay();
            networksOverlay.Layers.Add("Railways", railsLayer);

            ShapeFileFeatureLayer primaryLayer = new ShapeFileFeatureLayer(@"C:\CR\roads.shp",@"C:\CR\Primary.idx");
            primaryLayer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = LineStyles.Interstate4;
            primaryLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            networksOverlay.Layers.Add("PrimaryLayer", primaryLayer);
            winformsMap1.Overlays.Add("Networks", networksOverlay);
            // Define a Routing layer to render the route and stops
            
            routingSource = new RtgRoutingSource(@"C:\CR\roads.rtg");
            routingEngine = new RoutingEngine(routingSource, roadsSource);
            routingLayer = new RoutingLayer();
            LayerOverlay routingOverlay = new LayerOverlay();
            routingOverlay.Layers.Add("RoutingLayer", routingLayer);
            winformsMap1.Overlays.Add("RoutingOverlay", routingOverlay);
            pointLayer = new InMemoryFeatureLayer();
            pointLayer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle = new PointStyle(PointSymbolType.Circle, new GeoSolidBrush(GeoColor.GetRandomGeoColor(RandomColorType.All)), 10);
            pointLayer.ZoomLevelSet.ZoomLevel01.DefaultTextStyle = TextStyles.Capital1("name");
            pointLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            LayerOverlay pointOverlay = new LayerOverlay();
            pointOverlay.Layers.Add("PointLayer", pointLayer);
            winformsMap1.Overlays.Add("PointOverlay", pointOverlay);
            labelCurrentScale.Text = String.Format("{0:f2}",winformsMap1.CurrentScale);
            labelMaximumScale.Text = String.Format("{0:f2}", winformsMap1.MaximumScale);
            labelMinimumScale.Text = String.Format("{0:f2}", winformsMap1.MinimumScale);
            railsLayer.Open();
            railsLayer.Close();
            winformsMap1.Refresh();
        }

        private void button1_Click(object sender, EventArgs e)
        {

        }

        private void button2_Click(object sender, EventArgs e)
        {
            double latitude = Convert.ToDouble(textLatitude.Text, CultureInfo.InvariantCulture);
            double longitude = Convert.ToDouble(textLongitude.Text, CultureInfo.InvariantCulture);
            Feature feature = new Feature(longitude, latitude, textName.Text);

            InMemoryFeatureLayer pointLayer = (InMemoryFeatureLayer)winformsMap1.FindFeatureLayer("PointLayer");
            if (!pointLayer.InternalFeatures.Contains(textName.Text))
            {
                pointLayer.InternalFeatures.Add(textName.Text, feature);
            }
            winformsMap1.Refresh(winformsMap1.Overlays["PointOverlay"]);
        }

        private void button3_Click(object sender, EventArgs e)
        {
            //RoutingLayer routingLayer = (RoutingLayer)((LayerOverlay)winformsMap1.Overlays["RoutingOverlay"]).Layers["RoutingLayer"];
            routingLayer.Open();
            routingLayer.Routes.Clear();
            routingLayer.StartPoint = new PointShape(17.97,49.47);
            routingLayer.EndPoint = new PointShape(17.25,49.59);
            FeatureLayer streetLayer = (FeatureLayer)winformsMap1.FindFeatureLayer("Railways");
            PointShape closestPointStart = SnapToFeature(new Collection<FeatureLayer> { streetLayer }, new PointShape(17.97,49.47), GeographyUnit.DecimalDegree, 2500, DistanceUnit.Meter);
            PointShape closestPointEnd = SnapToFeature(new Collection<FeatureLayer> { streetLayer }, new PointShape(17.25,49.59), GeographyUnit.DecimalDegree, 2500, DistanceUnit.Meter);
            routingLayer.StartPoint = closestPointStart;
            routingLayer.EndPoint = closestPointEnd;
            RoutingResult rRes = routingEngine.GetRoute(closestPointStart,closestPointEnd);
            if (rRes != null)
            {
                routingLayer.Routes.Add(rRes.Route);
            }
            
            routingLayer.Close();
            winformsMap1.Refresh(winformsMap1.Overlays["RoutingOverlay"]);
        }

        private void winformsMap1_MouseMove(object sender, MouseEventArgs e)
        {
            //Displays the X and Y in screen coordinates.
            statusStrip1.Items["toolStripStatusLabelScreen"].Text = "X:" + e.X + " Y:" + e.Y;

            //Gets the PointShape in world coordinates from screen coordinates.
            PointShape pointShape = ExtentHelper.ToWorldCoordinate(winformsMap1.CurrentExtent, new ScreenPointF(e.X, e.Y), winformsMap1.Width, winformsMap1.Height);

            //Displays world coordinates.
            statusStrip1.Items["toolStripStatusLabelWorld"].Text = "(world) X:" + Math.Round(pointShape.X, 4) + " Y:" + Math.Round(pointShape.Y, 4);
        }

        private PointShape SnapToFeature(IEnumerable<FeatureLayer> snapToLayers, PointShape pointToSnap, GeographyUnit unitOfSnapToPoint, double maxDistance, DistanceUnit unitOfMaxDistance)
        {
            PointShape closestPoint = null;
            double closestDistance = double.MaxValue;

            // Loop through each layer to see which on have the closest point
            foreach (FeatureLayer featureLayer in snapToLayers)
            {
                // Make sure the layer is open
                featureLayer.Open();

                // Get all the features within our tolerance
                Collection<Feature> closeStreets = featureLayer.QueryTools.GetFeaturesInsideBoundingBox(new EllipseShape(pointToSnap, maxDistance, unitOfSnapToPoint, unitOfMaxDistance).GetBoundingBox(), ReturningColumnsType.NoColumns);

                // Loop through each feature in our tolerance
                foreach (Feature feature in closeStreets)
                {
                    // Find out the distance to the closest point on the feature
                    double snapDistance = pointToSnap.GetDistanceTo(feature.GetShape(), unitOfSnapToPoint, unitOfMaxDistance);

                    // If the distance is closer than the previous closest distance
                    // then this is the new winner so far
                    if (snapDistance < closestDistance)
                    {
                        // Make sure the distance is within our tolerance
                        if (snapDistance < maxDistance)
                        {
                            closestDistance = snapDistance;
                            closestPoint = feature.GetShape().GetClosestPointTo(pointToSnap, unitOfSnapToPoint);
                        }
                    }
                }
            }

            // return the closest point on the closes feature.  Note that this may return null 
            // if there are no features within the tolerance
            return closestPoint;
        }

        int pointsNumber = 1;

        private void winformsMap1_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Right)
            {
                PointShape pointShape = ExtentHelper.ToWorldCoordinate(winformsMap1.CurrentExtent, new ScreenPointF(e.X, e.Y), winformsMap1.Width, winformsMap1.Height);
                Feature feature = new Feature(pointShape.X, pointShape.Y,pointsNumber.ToString());
                pointLayer.InternalFeatures.Add(pointsNumber.ToString(), feature);
                pointsNumber++;
                winformsMap1.Refresh(winformsMap1.Overlays["PointOverlay"]);
            }
        }

        private void btnDM_Click(object sender, EventArgs e)
        {
            PointShape[] points = new PointShape[pointLayer.InternalFeatures.Count];
            int index = 0;
            foreach (Feature item in pointLayer.InternalFeatures)
            {
                points[index] = item.GetShape().GetCenterPoint();
                index++;
            }
            Stopwatch watch = new Stopwatch();
            watch.Start();
            RoutingResult[,] routeMatrix = new RoutingResult[points.Length, points.Length];
            for (int i = 0; i < points.Length; i++)
            {
                for (int j = i + 1; j < points.Length; j++)
                {
                    RoutingResult result = routingEngine.GetRoute(points[i], points[j]);
                    if (result.RouteSegments.Count <= 0)
                    {
                        result.Distance = double.MaxValue;
                    }
                    routeMatrix[i, j] = result;
                    // Reverse the result
                    RoutingResult reverseResult = new RoutingResult();
                    for (int k = result.Features.Count - 1; k >= 0; k--)
                    {
                        reverseResult.Features.Add(result.Features[k]);
                        reverseResult.Route.Lines.Add(result.Route.Lines[k]);
                        reverseResult.RouteSegments.Add(result.RouteSegments[k]);
                    }
                    reverseResult.Distance = result.Distance;
                    routeMatrix[j, i] = reverseResult;
                }
            }
            watch.Stop();
            textBoxRoutingTime.Text = watch.Elapsed.ToString();
        }
    }
}



